# Data Science for Business II | Week 2 Example 3: Stochastic production and inventory optimization
# Author: Lauri Neuvonen

# SETUP
# We'll use the Gurobi optimizer in this exercise, so we'll import it and Pandas first:

import pandas as pd
import numpy as np
from gurobipy import *

# DATA INPUT
# First, let's create the data table (without the quarters...
# as these match the indexes of the rows so they are not necessary):


# The following creates a 4x4 data matrix

productionData = np.array([[700, 2, 0.25],
                            [400, 5, 0.25],
                            [500, 3, 0.25]])

# The following creates a name vector for the columns in the matrix:
columnNames = ("Production capacity (m2)", "Production cost (EUR/m2)", "Inventory cost (EUR/m2)")

# In Python, we can create a data frame object by using the Pandas library:
# it's usually a good idea to keep the data in a "data frame" object.
# Let's use the previous objects to create a dataframe for the data
df = pd.DataFrame(productionData, columns=columnNames)

# CONSTRUCTING THE DECISION TREE
# We can use probability data from a file to generate the scenario tree:
pData = pd.read_csv("Example3-clean.csv", delimiter=";", index_col = 0, decimal=",")

# These variables record the indices of lines corresponding to different levels
l1Ind = pData.index[pData["level"] == 1] # indices of level 1 rows
l2Ind = pData.index[pData["level"] == 2]
l3Ind = pData.index[pData["level"] == 3]
l4Ind = pData.index[pData["level"] == 4]

# Selection of probabilities and Demand volumes from data
probabilities = pData["p"]
demand = pData["Demand"]

# BUILDING THE OPTIMIZATION MODEL
# Now that we have the data in a dataframe, let's build the optimization model

# From the LP formulation, we see that the constraint matrix A has 8 rows (constraints) and 8 columns (variables)

# Numbers of decision variables in different quarters
N_variations = 3 # number of variations. Assumed same number for each level in tree.
N_Q1vars = 1
N_Q2vars = N_variations*N_Q1vars
N_Q3vars = N_variations*N_Q2vars

fP = df["Production cost (EUR/m2)"]
fI = df["Inventory cost (EUR/m2)"]

print(pData)

try:

    # Let's first create an "empty" LP model:
    m = Model("Example 3 Optimizer")

    # Production variables: one set per each quarter
    xQ1 = m.addVar()
    xQ2 = m.addVars(range(N_Q2vars))
    xQ3 = m.addVars(range(N_variations), range(N_variations))

    # Inventory variables: one set per each quarter
    sQ1 = m.addVar()
    sQ2 = m.addVars(range(N_Q2vars))
    sQ3 = m.addVars(range(N_variations), range(N_variations))

    m.update() # required to make variable queries work

    demand = demand.values
    probabilities = probabilities.values

    # As Gurobi works nicely with dictionaries, we'll form probability dictionaries for Q2 and Q3:
    l2probs = probabilities[l2Ind] # picks the correct rows from the tree's probability column
    l2paramKeys = xQ2.keys() # finds the group of keys used to access variables xQ2
    l2Pcoeff = dict(zip(l2paramKeys, l2probs)) # forms a dictionary that maps xQ2 keys to matching probabilities. Note: the probabilities are in the order matching xQ2's

    # same for Q3, this time we however need to include the correspondin level 2 probabilities to get the final probability coefficients
    l3probs = probabilities[l3Ind].reshape(N_variations, N_variations) # reforms the level 3 probabilities as a matrix: rows correspond to l2 scenarios
    l3scenprobs = l3probs*l2probs[:, np.newaxis] # multiply level probabilities to get the scenario probabilities. "newaxis" is a trick to help with calculation
    l3scenprobs = l3scenprobs.flatten() # back to a vector
    l3paramKeys = xQ3.keys() # get the keys to use in the dictionary
    l3Pcoeff = dict(zip(l3paramKeys, l3scenprobs)) # create the dictionary



    # Then, we'll set the constraints into the model:
    # In this case we use Gurobi's handy "addConstrs" method.

    # The production capacity constraints. Note python indexing starting from 0
    m.addConstr(xQ1 <= df["Production capacity (m2)"][0])
    m.addConstrs(xQ2[i] <= df["Production capacity (m2)"][1] for i in range(N_Q2vars))
    m.addConstrs(xQ3[i,j] <= df["Production capacity (m2)"][2] for i in range(N_variations) for j in range(N_variations))

    # Demand constraints: production + inventory from last week - inventory left from this week must equal demand
    m.addConstr(xQ1 - sQ1 == np.sum(demand[l2Ind].dot(probabilities[l2Ind])))
    m.addConstrs(xQ1 + xQ2[i] - sQ2[i] ==
                 demand[l2Ind][i] +
                 sum(demand[l3Ind][i*N_Q2vars+j]*(probabilities[l3Ind][i*N_Q2vars+j])
                     for j in range(N_variations))
                 for i in range(N_variations)) # j goes over 3rd level options, happens to match N_Q2vars
    m.addConstrs(xQ1 + xQ2[i] + xQ3[i, j] - sQ3[i,j] ==
                 demand[l2Ind][i] + demand[l3Ind][i*N_Q2vars+j] +
                 sum(demand[l4Ind][(i*N_variations+j)*N_variations + k] * probabilities[l4Ind][(i*N_variations+j)*N_variations + k]
                     for k in range(N_variations))
                 for i in range(N_variations)
                 for j in range(N_variations))


    # Objective function sums products of cost coefficients (fP & fI) and decision variables (x & s) :
    o1 = df["Production cost (EUR/m2)"][0] * xQ1
    o2 = df["Production cost (EUR/m2)"][1] * xQ2.prod(l2Pcoeff)
    o3 = df["Production cost (EUR/m2)"][2] * xQ3.prod(l3Pcoeff)
    o4 = df["Inventory cost (EUR/m2)"][0] * sQ1
    o5 = df["Inventory cost (EUR/m2)"][1] * sQ2.prod(l2Pcoeff)
    o6 = df["Inventory cost (EUR/m2)"][2] * sQ3.prod(l3Pcoeff)
    m.setObjective(o1 + o2 + o3 + o4 + o5 + o6, GRB.MINIMIZE)

    # Solve the problem:
    m.optimize()

    # Write model to a file (can be accessed with any text editor)
    m.write('Ex3model.lp')

    # get results into var_values
    var_values = m.getVars()

    for v in var_values:
        print('%s %g' % (v.varName, v.x))

    print('Obj: %g' % m.objVal)

except GurobiError as e:
    print('Error code ' + str(e.errno) + ": " + str(e))

except AttributeError:
    print('Encountered an attribute error')